


#include "trigonometric.h"


int is_spheres_overlap( sphere_t* sphere_v1, sphere_t* sphere_v2)
{
	DIST_TYPE dist;

	dist = calc_dist( sphere_v1->center, sphere_v2->center);

	if( dist <= sphere_v1->radius + sphere_v2->radius)
		return 1;
	else
		return 0;
}

int is_points_overlap( point_t* point_v1, point_t* point_v2)
{
	int i, dim;

	dim = point_v1->dim;

	for( i=0; i<dim; i++)
	{
		if( abs( point_v1->coord[ i], point_v2->coord[ i]) > PRECISION_THR)
			return 0;
	}

	return 1;
}

point_t* cons_vector( point_t* point_v1, point_t* point_v2)
{
	int i, dim;
	point_t* point_v;

	dim = point_v1->dim;

	point_v = alloc_point( dim);

	for( i=0; i<dim; i++)
	{
		point_v->coord[ i] = point_v2->coord[ i] - point_v1->coord[ i];
	}
	
	return point_v;
}

DIST_TYPE calc_dot_product( point_t* vector_v1, point_t* vector_v2)
{
	int i, dim;
	DIST_TYPE dot_product;

	dim = vector_v1->dim;
	dot_product = 0;
	for( i=0; i<dim; i++)
	{
		dot_product += vector_v1->coord[ i] * vector_v2->coord[ i];
	}

	return dot_product;
}

DIST_TYPE calc_length_vector( point_t* vector_v)
{
	int i, dim;
	DIST_TYPE leng_v;

	dim = vector_v->dim;
	leng_v = 0;
	for( i=0; i<dim; i++)
	{
		leng_v += ( DIST_TYPE)pow( vector_v->coord[ i], 2);
	}

	return ( DIST_TYPE)sqrt( leng_v);
}
	
double calc_angle_vectors( point_t* vector_v1, point_t* vector_v2)
{
	DIST_TYPE leng_v1, leng_v2, dot_product;

	dot_product = calc_dot_product( vector_v1, vector_v2);

	leng_v1 = calc_length_vector( vector_v1);
	
	leng_v2 = calc_length_vector( vector_v2);

	if( leng_v1 == 0 || leng_v2 == 0)
	{
		fprintf( stderr, "Invalid vectors [calc_angle_vectors].\n");
		exit( 0);
	}

	return acos( dot_product / ( leng_v1 * leng_v2));
}


DIST_TYPE calc_dist_phi( sphere_t* s_q, point_t* point_v, double phi)
{
	DIST_TYPE r_q, dist_qv, dist_phi;

	r_q = s_q->radius;
	dist_qv = calc_dist( s_q->center, point_v);

	dist_phi = ( DIST_TYPE)sqrt( r_q * r_q + dist_qv * dist_qv - 2 * r_q * dist_qv * cos( phi)); 
	
	return dist_phi;
}

DIST_TYPE calc_g_function( sphere_t* s_q, sphere_t* s_a, sphere_t* s_b, double phi, double delta)
{
	DIST_TYPE minDist, maxDist, dist_phi, dist_phi_minus_delta;

	dist_phi = calc_dist_phi( s_q, s_a->center, phi);
	dist_phi_minus_delta = calc_dist_phi( s_q, s_b->center, phi - delta);

	maxDist = dist_phi + s_a->radius;
	minDist = dist_phi_minus_delta - s_b->radius;

	return maxDist - minDist;
}

int is_dists_equal( DIST_TYPE dist_v1, DIST_TYPE dist_v2)
{
	if( abs( dist_v1, dist_v2) <= DIST_PRECISION_THR)
		return 1;
	else
		return 0;
}

int trigonometric_pruning( sphere_t* s_a, sphere_t* s_b, sphere_t* s_q)//, int pre_check_tag)
{
	int pre_check_tag;
	DIST_TYPE dist_qa, dist_qb, max_dist, min_dist;
	double delta, phi_1, phi_2;
	point_t* vector_qa, *vector_qb;

	//emp
	dup_point( s_a->center);
	dup_point( s_b->center);
	dup_point( s_q->center);
	//emp

	pre_check_tag = PRE_CHECK_TAG;

	//Pre-check whether S_a and S_b overlap.
	if( pre_check_tag == 1)
	{
		if( is_spheres_overlap( s_a, s_b))
			return 0;
	}

	//Check whether S_b and S_q overlap.
	if( is_spheres_overlap( s_b, s_q))
		return 0;

	dist_qa = calc_dist( s_q->center, s_a->center);
	dist_qb = calc_dist( s_q->center, s_b->center);

	//Check whether c_a happens to be c_q;
	if( is_points_overlap( s_a->center, s_q->center))
	{
		//In this case, maxDist = r_q + r_a; minDist = dist_qb - r_q - r_b;
		min_dist = s_q->radius + s_a->radius;
		max_dist = dist_qb - s_q->radius - s_b->radius;

		if( min_dist < max_dist)
			return 1;
		else
			return 0;
	}
	else
	{
		//In this case, c_a does not overlap with c_q;
		//Compute delta;
		vector_qa = cons_vector( s_q->center, s_a->center);
		vector_qb = cons_vector( s_q->center, s_b->center);

		delta = calc_angle_vectors( vector_qa, vector_qb);

		//Check whether dist_qa == dist_qb * cos( delta);
		//Which is equvilant to whether g' has two roots of phi_1 = 1/2 PI and phi_2 = 3/2 PI.
		if( is_dists_equal( dist_qa, dist_qb * cos( delta)))
		{
			phi_1 = 0.5 * PI;
			phi_2 = 1.5 * PI;
		}
		else
		{
			//the other case;
			//we utilize the the results in the paper.
			phi_1 = ( double)atan( ( dist_qb * sin( delta)) / ( dist_qb * cos( delta) * dist_qa));
			phi_2 = phi_1 + PI;
		}

		//Release the resources.
		release_point( vector_qa);
		release_point( vector_qb);

		//Check whether it holds that g( phi_1) < 0 and g( phi_2) < 0;
		if( calc_g_function( s_q, s_a, s_b, phi_1, delta) < 0 &&
			calc_g_function( s_q, s_a, s_b, phi_2, delta) < 0)
			return 1;
		else
			return 0;
	}
}
